#r "nuget: arcIO.NET, 0.1.0-preview.7"
#r "nuget: ISADotNet.QueryModel, 0.7.0-preview.10"
#r "nuget: BioFSharp.IO, 2.0.0-preview.3"
#r "nuget: FSharp.Stats, 0.4.12-preview.1"
#r "nuget: Plotly.NET, 3.0.1"
#r "nuget: Deedle"
#r "nuget: DeedleExtensions, 0.2.0"
#r "nuget: metaXa, 0.1.0"
#r "nuget: Argu,6.1.1"
open MetaXa
open MetaFrame
open Stats
open FSharp.Stats
open Plotly.NET
open Deedle
open BioFSharp.IO
open DeedleExtensions
open ISADotNet
open ISADotNet
open ISADotNet.QueryModel

module Annotate =

    open System
    open Argu
    open Deedle 

    printfn "=====\nCheck if dotnet sdk reference can be rechanged to latest: https://stackoverflow.com/a/65934809 \n====="

    /// \t or "\t" doesnt work as argument
    let getSeparator str = 
        match str with
        | "tab"             -> "\t"
        | "tabulator"       -> "\t"
        | _ -> str

    /// Snapshots of MapMan and Gene Ontology were generated with FATool and genome releases 5.5 (Chlamy) and Araport11 (Arabidopsis)
    module Mapping =
    
        /// reads FA tool snapshot for araport11 or chlamyJGI_v5.5 or other annotation files
        let readMapping (mappingFilePath:string) (columnSeparator:string) (identifierColHeader:string) :Frame<string,string> = 
            Frame.ReadCsv(mappingFilePath,hasHeaders=true,separators = columnSeparator)
            |> Frame.indexRows identifierColHeader
 
        /// truncateID processes identifier (e.g. Cre10.g123456.t2.1 -> Cre10.g123456)
        let getAnnotationsFromIdentifier (frame:Frame<string,string>) (annotationHeader:string[]) (multipleIdentifierSeparator:string) (multipleAnnotationSeparator:string) (truncateID:string->string) (identifier:string) =
        
            /// identifier that should be mapped
            let identifier = identifier.Split([|multipleIdentifierSeparator|],StringSplitOptions.None)
            
            /// Mappings from all identifer to annotations, that are present in annotation frame
            let mappings :Series<string,string> []= 
                identifier 
                |> Array.choose (fun ident -> 
                    let truncId = truncateID ident
                    let k = frame.TryGetRow truncId
                    if k.HasValue then 
                        Some k.Value
                    else
                        printfn "Warning: The following id could not be found within mapping file: %s" truncId
                        None
                    )

            /// all annotations that should be used
            annotationHeader //["GO","Synonym"]
            |> Array.map (fun annotationType -> 
                mappings //[[GO => GO:006; Synonym => Q0WV96];[GO => GO:001; Synonym => Q01337]]
                |> Array.map (fun mapping -> 
                    let annotation = mapping.[annotationType]
                    annotation.Split ';' 
                    |> String.concat multipleAnnotationSeparator
                    )
                )

        let getAnnotationRow (frame:Frame<string,string>) (annotationHeader:string[]) multipleIdentifierSeparator multipleAnnotationSeparator truncateID (identifier:string) = 
            getAnnotationsFromIdentifier frame annotationHeader multipleIdentifierSeparator multipleAnnotationSeparator truncateID identifier
            |> Array.map (String.concat multipleIdentifierSeparator)


    /// User data is read, annotated and written to a new file
    module Data = 
    
        let getDataFrame (columnSeparator:string) inputPath = 
            System.IO.File.ReadAllLines(inputPath)
            |> Array.map (fun x ->
                x.Split([|columnSeparator|],System.StringSplitOptions.None)
                )

        /// index of column that contains the identifier to annotate
        let getColIndex (dataFrame:string[][]) columnHeader= 
            Array.tryFindIndex (fun x -> x = columnHeader) dataFrame.[0]
            |> fun o -> 
                match o with
                | Some i -> i
                | _ -> failwithf "ColumnHeader %s not found." columnHeader
    
        /// based on given mapping arguments the file is extended with given mapping columns
        let getHeader (dataFrame:string[][]) (columnSeparator:string) (annotationHeader:string[]) = 
            Array.append dataFrame.[0] annotationHeader
            |> String.concat columnSeparator
   
        /// every row of the file is processed and converted to a new string with additional information attached at the end of the line
        let getAnnotatedLines annotationFrame inputPath (columnSeparator:string) columnHeader (annotations:string[]) (multipleIdentifierSeparator:string) truncateID (multipleAnnotationSeparator:string) =       
            let dataFrame = getDataFrame columnSeparator inputPath           
            let colIndex = getColIndex dataFrame columnHeader     
            let header = getHeader dataFrame columnSeparator annotations
            //let rowCount = dataFrame.Length
            dataFrame
            |> Array.tail
            |> Array.mapi (fun i x -> 
                let identifier = x.[colIndex].Replace("\"","")
            
                //if i%50=0 then printfn "%04i/%i: %s" i rowCount identifier
                //printfn "%04i/%i: %s" i rowCount identifier
                let annotations = 
                    if identifier = "" then 
                        Array.init annotations.Length (fun _ -> "")
                    else 
                        Mapping.getAnnotationRow annotationFrame annotations multipleIdentifierSeparator multipleAnnotationSeparator truncateID identifier
                Seq.append x annotations
                |> String.concat columnSeparator
                )
            |> Array.append [|header|]

        let annotateAndWriteData annotationFrame inputPath (columnSeparator:string) columnHeader truncateID (annotations:string[]) (multipleIdentifierSeparator:string) (multipleAnnotationSeparator:string) outputPath =
            /// warns user if separators are identical
            let checkSeparators = 
                if columnSeparator = multipleIdentifierSeparator then failwith "WARNING: Column separator is equal to identifier separator"
                if columnSeparator = multipleAnnotationSeparator then failwith "WARNING: Column separator is equal to annotation separator"
            let annotatedRows = 
                getAnnotatedLines annotationFrame inputPath columnSeparator columnHeader annotations multipleIdentifierSeparator truncateID multipleAnnotationSeparator
            System.IO.File.WriteAllLines(outputPath,annotatedRows)

module SamPlotting =
    open Deedle
    open FSharp.Stats
    open Plotly.NET
    open FSharp.Stats.Testing
    open SAM
    open Plotly.NET.StyleParam

    let createSAMChart res = 

        let observed = [| res.NegSigBioitem; res.NonSigBioitem; res.PosSigBioitem|] |> Array.concat 
        let obs = observed |> Array.map (fun x -> x.Statistics) 
        let expected = res.AveragePermutations |> Array.map (fun x -> x.Statistics)
        let minDi = Seq.min obs
        let maxDi = Seq.max obs


        // positive significant changes 
        let posExpected = expected.[res.NegSigBioitem.Length + res.NonSigBioitem.Length .. res.NegSigBioitem.Length + res.NonSigBioitem.Length + res.PosSigBioitem.Length-1]
        let posChart = 
            Chart.Point(posExpected,res.PosSigBioitem |> Array.map (fun x -> x.Statistics))
            |> Chart.withLineStyle(Color=Color.fromKeyword Green)
            |> Chart.withTraceInfo("positive change",Visible = Visible.True )


        // no significant changes
        let nonex = expected.[res.NegSigBioitem.Length .. res.NegSigBioitem.Length + res.NonSigBioitem.Length-1]
        let nonchart = 
            Chart.Point(nonex,res.NonSigBioitem |> Array.map (fun x -> x.Statistics))
            |> Chart.withLineStyle(Color=Color.fromKeyword Gray)
            |> Chart.withTraceInfo("no change",Visible = Visible.True)

        // negative significant changes 
        let negex = expected.[0 .. res.NegSigBioitem.Length-1]
        let negchart = 
            Chart.Point(negex,res.NegSigBioitem |> Array.map (fun x -> x.Statistics))
            |> Chart.withLineStyle(Color=Color.fromKeyword Red)
            |> Chart.withTraceInfo("negative change",Visible = Visible.True)

        let samValues = 
            [
                negchart
                nonchart
                posChart
            ]
            |> Chart.combine
        
        let chartConfig =
            let svdConfig =
                ConfigObjects.ToImageButtonOptions.init(
                    Format = StyleParam.ImageFormat.SVG)
            Config.init (
                ToImageButtonOptions = svdConfig,
                ModeBarButtonsToAdd=[ModeBarButton.HoverCompareCartesian]
            
            )

        let cutLineUp = [(minDi + res.Delta) ; (maxDi + res.Delta)]
        let cutsUp = 
            Chart.Line(cutLineUp,[minDi;maxDi])
            |> Chart.withLineStyle(Color=Color.fromKeyword Purple,Dash = StyleParam.DrawingStyle.Dash, Width = 0.5)
            |> Chart.withTraceInfo("delta",Visible = Visible.True)
        let cutLineLow = [(minDi - res.Delta) ; (maxDi - res.Delta)]
        let cutsLow = 
            Chart.Line(cutLineLow,[minDi;maxDi])
            |> Chart.withLineStyle(Color=Color.fromKeyword Purple,Dash = StyleParam.DrawingStyle.Dash, Width = 0.5)
            |> Chart.withTraceInfo("delta",Visible = Visible.True)
        let linechart = 
            Chart.Line([minDi;maxDi], [minDi;maxDi])
            |> Chart.withTraceInfo("bisecting angle",Visible = Visible.True)
            |> Chart.withLineStyle(Color=Color.fromKeyword Black, Width = 1)

        let uppercut = 
            let xAnchorUppercut = [minDi .. 5. .. maxDi]
            Chart.Line (xAnchorUppercut, List.init xAnchorUppercut.Length (fun x -> res.UpperCut))

            |> Chart.withTraceInfo("upper cut",Visible = Visible.True)
            |> Chart.withLineStyle(Color=Color.fromKeyword Black,Dash = StyleParam.DrawingStyle.Dash, Width = 0.3)

        let lowercut = 
            Chart.Line([minDi;maxDi],[res.LowerCut;res.LowerCut])
            |> Chart.withTraceInfo("lower cut",Visible = Visible.True)
            |> Chart.withLineStyle(Color=Color.fromKeyword Black,Dash = StyleParam.DrawingStyle.Dash,Width = 0.3)
            //|> Chart.withXAxisStyle(MinMax = (-15,20))
        
            |> Chart.withTraceInfo("lower cut",Visible = Visible.True)
        let plot = 
            [linechart;
            samValues;
            cutsUp;
            cutsLow;
            uppercut;
            lowercut]
            |> Chart.combine
            |> Chart.withTitle(title = "SAM results")
            |> Chart.withXAxisStyle("expected Score")
            |> Chart.withYAxisStyle ("observed Score")
            |> Chart.withConfig(chartConfig)
            |> Chart.withTemplate(ChartTemplates.lightMirrored)
        plot 

module Plotting =

    open FSharp.Stats
    open Plotly.NET
    open Deedle
    open BioFSharp.IO
    open DeedleExtensions

    let plotMA highlightList (f:Frame<string,string>) = 
        let dataAll = 
            f
            |> Frame.mapRows (fun k s -> 
                let a = s.GetAs<float>("MeanAbundance",nan)
                let r = s.GetAs<float>("MeanR",nan)
                let stDev = s.GetAs<float>("StDevR",nan)
                let l = 
                    let x = s.GetAs<string>("ProtName","")
                    if x = "" then 
                        "" 
                    else 
                        x
                let c = s.GetAs<bool>("IsCandidate")
                {|
                        ProtName         = l
                        MeanAbundance    = a
                        MeanR            = r
                        StDevR           = stDev
                        IsCandidate      = c
                |}
                )
            |> Series.values
            |> Array.ofSeq
    
        let all = 
            let x = dataAll |> Seq.map (fun x -> x.MeanR) 
            let y = dataAll |> Seq.map (fun x -> x.MeanAbundance) 
            let label = dataAll |> Seq.map (fun x -> x.ProtName)
            Chart.Point(y,x, MultiText = label)

        let candidates =
            let dataFiltered = dataAll |> Array.filter (fun x -> x.IsCandidate)  
            let x = dataFiltered |> Seq.map (fun x -> x.MeanR) 
            let y = dataFiltered |> Seq.map (fun x -> x.MeanAbundance) 
            let labels = dataFiltered |> Seq.map (fun x -> x.ProtName)
            Chart.Point(
            y,x,
            Name="points",
            MultiText=labels,
            TextPosition=StyleParam.TextPosition.TopRight
            )
            |> Chart.withTraceInfo "candidates"
        let candidatesNoStabw =
            let dataFiltered = dataAll |> Array.filter (fun x -> x.IsCandidate)  
            let x = dataFiltered |> Seq.map (fun x -> x.MeanR) 
            let y = dataFiltered |> Seq.map (fun x -> x.MeanAbundance) 
            let labels = dataFiltered |> Seq.map (fun x -> x.ProtName)
        
            Chart.Point(
            y,x,
            Name="points", MultiText = labels
            )
            |> Chart.withTraceInfo "candidatesNoStabw"
        let candidates' =
            let dataFiltered = dataAll |> Array.filter (fun x -> highlightList |> List.contains x.ProtName )  
            let x = dataFiltered |> Seq.map (fun x -> x.MeanR) 
            let y = dataFiltered |> Seq.map (fun x -> x.MeanAbundance) 
            let labels = dataFiltered |> Seq.map (fun x -> x.ProtName)
            Chart.Point(
            y,x,
            Name="points",
            MultiText=labels,
            TextPosition=StyleParam.TextPosition.TopRight
            )
            |> Chart.withTraceInfo "candidates Reci"
        [
            all
            candidatesNoStabw
            candidates'

        ]
        |> Chart.combine
        |> Chart.withYAxisStyle "log2(bait)-log2(control)"
        |> Chart.withXAxisStyle "log2(Mean intensity)"

    let plotVulcano highlightList f = 
        let dataAll = 
            f
            |> Frame.mapRows (fun k s -> 
                let a = s.GetAs<float>("MeanAbundance",nan)
                let r = s.GetAs<float>("MeanR",nan)
                let stDev = s.GetAs<float>("StDevR",nan)
                let qValue = s.GetAs<float>("qValue",nan)
                let l = 
                    let x = s.GetAs<string>("ProtName","")
                    if x = "" then 
                        "" 
                    else 
                        x
                let c = s.GetAs<bool>("IsCandidate")
                {|
                        ProtName         = l
                        MeanAbundance    = a
                        MeanR            = r
                        StDevR           = stDev
                        IsCandidate      = c
                        QVal = qValue
                |}
                )
            |> Series.values
            |> Array.ofSeq
    
        let all = 
            let y = dataAll |> Seq.map (fun x -> x.MeanR)
            let x = 
                dataAll 
                |> Seq.map (fun x -> x.QVal) 
                |> Seq.map (fun x -> 
                    let res = -log10 x
                    if infinity.Equals res then 4. else res
                    )
            let label = dataAll |> Seq.map (fun x -> x.ProtName)
            let xError = 
                dataAll |> Seq.map (fun x -> x.StDevR)
            Chart.Point(y,x, MultiText = label)
        let candidatesNoStabw =
            let dataFiltered = dataAll |> Array.filter (fun x -> x.IsCandidate)  
            let y = dataFiltered |> Seq.map (fun x -> x.MeanR) 
            let x = 
                dataFiltered 
                |> Seq.map (fun x -> x.QVal) 
                |> Seq.map (fun x -> 
                    let res = -log10 x
                    if infinity.Equals res then 4. else res
                    )
            let labels = dataFiltered |> Seq.map (fun x -> x.ProtName)
            let xError = dataFiltered |> Seq.map (fun x -> x.StDevR)
            Chart.Point(
            y,x,
            Name="points",
            MultiText=labels
            )
            |> Chart.withTraceInfo "candidatesNoStabw"
        let candidates' =
            let dataFiltered = dataAll |> Array.filter (fun x -> highlightList |> List.contains x.ProtName )  
            let y = dataFiltered |> Seq.map (fun x -> x.MeanR) 
            let x = 
                dataFiltered 
                |> Seq.map (fun x -> x.QVal) 
                |> Seq.map (fun x -> 
                    let res = -log10 x
                    if infinity.Equals res then 4. else res
                    )
            let labels = dataFiltered |> Seq.map (fun x -> x.ProtName)
            Chart.Point(
            y,x,
            Name="points",
            MultiText=labels,
            TextPosition=StyleParam.TextPosition.TopRight
            )
            |> Chart.withTraceInfo "candidates Reci"
        [
            all
            candidatesNoStabw
            candidates'

        ]
        |> Chart.combine
        |> Chart.withYAxisStyle "log2(bait)-log2(control)"
        |> Chart.withXAxisStyle "log2(Mean intensity)"

    let plotHisto (f:Frame<string,string>) =
        let dataAll = 
            f
            |> Frame.mapRows (fun k s -> 
                let a = s.GetAs<float>("MeanAbundance",nan)
                let r = s.GetAs<float>("MeanR",nan)
                let stDev = s.GetAs<float>("StDevR",nan)
                let l = ""//s.GetAs<string>("ProtName","")
                let c = s.GetAs<bool>("IsCandidate")
                {|
                        ProtName         = l
                        MeanAbundance    = a
                        MeanR            = r
                        StDevR           = stDev
                        IsCandidate      = c
                |}
                )
            |> Series.values
            |> Array.ofSeq
        let histo =     
            let data = dataAll |> Seq.map (fun x -> x.MeanR)   |> Seq.filter (fun x -> nan.Equals x |> not)|> Array.ofSeq
            let bw = FSharp.Stats.Distributions.Bandwidth.nrd0 data 
            let binned = FSharp.Stats.Distributions.Frequency.create  (bw) data |> Map.toSeq
            Chart.Column binned
            |> Chart.withXAxisStyle "log2(bait)-log2(control)"
        histo
    
module MetaxaTemp =
    open MetaXa
    open MetaFrame
    open Stats
    open FSharp.Stats
    open Plotly.NET
    open Deedle
    open BioFSharp.IO
    open DeedleExtensions
    open ISADotNet


    module Global = 

        /// As used by Deseq2, see: https://github.com/hbctraining/DGE_workshop/blob/master/lessons/02_DGE_count_normalization.md 
        module MedianOfRatios = 

            /// expects data frame with shape frame<sampleIds,GeneIds>
            let medianOfRatiosWide (data:Frame<string,string>) =
                //data frame with shape frame<sampleIds,GeneIds>
                let ratios = 
                    data
                    |> Frame.dropSparseCols
                    |> Frame.getNumericCols
                    |> Series.map (fun geneID s -> 
                        let geometricMeanAcrossAllSamples = 
                            s.Values 
                            |> FSharp.Stats.Seq.median
                        let ratioPerSample = 
                            s
                            |> Series.map (fun x sample ->  sample / geometricMeanAcrossAllSamples)
                        ratioPerSample
                        )
                    |> Frame.ofColumns
                // Series of shape: Series<sampleId,correctionFactor:float>
                let correctionFactorPerSample = 
                    ratios
                    |> Frame.transpose
                    |> Stats.median
                // data frame with frame<sampleIds,GeneIds>
                // contains gene data divided by the sample wise estimated correction factor
                let correctedData = 
                    data
                    |> Frame.getNumericCols 
                    |> Series.map (fun geneID s -> 
                        s / correctionFactorPerSample
                        )
                    |> Frame.ofColumns
                let abundanceInit = 
                    Chart.Column(data |> Frame.transpose |> Frame.dropSparseRows |> Stats.median |> Series.observations)
                    |> Chart.withXAxisStyle "SampleId"
                    |> Chart.withXAxisStyle "Median of Initial Abundances" 
                let c = 
                    Chart.Column(correctionFactorPerSample |> Series.observations)
                    |> Chart.withXAxisStyle "SampleId"
                    |> Chart.withXAxisStyle "Correction Factor."
                let abundanceCorrected = 
                    Chart.Column(correctedData |> Frame.transpose |> Frame.dropSparseRows  |> Stats.median |> Series.observations)
                    |> Chart.withXAxisStyle "SampleId"
                    |> Chart.withXAxisStyle "Median of corrected abundances" 
                correctedData,abundanceInit,c,abundanceCorrected


    module Imputation = 


        module SimpleSampling = 
            /// Expects normalized data.
            /// (learnVarianceMeanDependence: seq<seq<float>> -> (float -> float)) function, which learns how to map from a mean expression to an estimator for variance.
            let computeColumnWiseUsing (sampleFunction: float -> float -> float) minCountsForVarEst (quantileForLowMean:float option) (learnVarianceMeanDependence: seq<float*float> -> (float -> float)) (grouped:Frame<string*string,string>) =
                ///
                let meanWithAtLeast k (x:Series<_,float>) =
                    let tmp = Series.values x
                    if tmp |> Seq.length < k  then 
                        nan
                    else 
                        x |> Series.values |> Seq.mean 
                ///
                let cvWithAtLeast k (x:Series<_,float>) =
                    let tmp = Series.values x
                    if tmp |> Seq.length < k  then 
                        nan
                    else 
                        x |> Series.values |> cv //Seq.disp

                /// Computes means and variances of complete replicates.
                let log2MeanToCVEmpirical = 
                    grouped
                    |> Frame.getNumericCols
                    |> Series.map (fun geneID s -> 
                        let means = s |> (meanWithAtLeast minCountsForVarEst)
                        let cols = s |> (cvWithAtLeast minCountsForVarEst)
                        means, cols
                        )
                    |> Series.values
                    |> Array.ofSeq
                    |> Array.filter (fun x -> nan.Equals(fst x) |> not && nan.Equals(snd x) |> not)
                    |> Array.sortBy fst
                    |> Array.map (fun (mean,dis) -> log2 mean, dis)
                    |> fun data -> 
                        let borders = FSharp.Stats.Testing.Outliers.tukey 1.5 (data |> Array.map snd)
                        data 
                        |> Array.filter (fun x -> snd x > borders.Lower && snd x < borders.Upper )
                /// function mapping from a mean expression to an estimator for dispersion (cv).
                let log2MeanToCv = 
                    learnVarianceMeanDependence log2MeanToCVEmpirical

                let log2MeanToCVByFit = 
                    log2MeanToCVEmpirical
                    |> Array.map (fun x -> fst x, fst x |> log2MeanToCv)
                let meanToCvEstChart = 
                    [
                        [
                        
                            Chart.PointDensity(log2MeanToCVEmpirical |> Array.map fst,log2MeanToCVEmpirical |> Array.map snd)
                            // |> Chart.withMarkerStyle (3)
                        ]
                        |> Chart.combine
                        |> Chart.withTraceInfo "mean to CV empirical"
        
                        Chart.Spline(log2MeanToCVByFit)
                        |> Chart.withTraceInfo "mean Vs. CV fitted"
                    ]
                    |> Chart.combine
                    |> Chart.withXAxisStyle "log2(mean Abundance)"
                    |> Chart.withYAxisStyle "CV"

                let quantileForLowMean, quantilesToMeansChart = 
                    let data = 
                        Array.map fst log2MeanToCVEmpirical
                        |> Array.map (fun x -> 2.**x) 
                    match quantileForLowMean with 
                    | None -> 
                        None,
                        [
                            [|0.05 .. 0.05 .. 0.95|]
                            |> Array.map (fun q -> q,FSharp.Stats.Quantile.compute q data)
                            |> Chart.Point
                            |> Chart.withTraceInfo "quantiles VS Abundance"
                        ]
                        |> Chart.combine
                        |> Chart.withXAxisStyle "Quantiles"
                        |> Chart.withYAxisStyle "Abundance"
                    | Some quantileForLowMean -> 
                        let q = data |> FSharp.Stats.Quantile.compute quantileForLowMean
                        Some q,
                        [
                            [|0.00 .. 0.05 .. 0.95|]
                            |> Array.mapi (fun i q -> if i = 0 then 0.01 else q)
                            |> Array.map (fun q -> q,FSharp.Stats.Quantile.compute q data)
                            |> Chart.Point
                            |> Chart.withTraceInfo "quantiles VS Abundance"
                            Chart.Point([quantileForLowMean],[q])
                            |> Chart.withTraceInfo "selected quantile"
                        ]
                        |> Chart.combine
                        |> Chart.withXAxisStyle "Quantiles"
                        |> Chart.withYAxisStyle "Abundance"

                /// Imputation by sampling from a gausian normal distribution based on the input vector
                /// function is truncated at zero. Can not return negative values.
                let initImputeUsing sampleF (estimateCv:float -> float) (lowMean:float option) (d:  seq<float>) =          
                        let clean = d |> Seq.filter (fun x -> nan.Equals x |> not)
                        let sampleSafe mean standardDev = 
                            let rec sample () = 
                                let tmp = sampleF mean standardDev 
                                if tmp < 0. then sample () else tmp
                            sample()
                        if clean |> Seq.isEmpty |> not then 
                            let mean = Seq.mean clean
                            let cvEst = estimateCv mean 
                            let standardDev = cvEst * mean 
                            //printfn "mean %f" mean
                            //printfn "stndev %f" standardDev
                            //printfn "before %A" d 
                            let res = 
                                d 
                                |> Seq.map (fun x -> if nan.Equals x then sampleSafe mean standardDev else x)
                            //printfn "%A" res
                            res
                        else 
                            match lowMean with 
                            | Some lowMean -> 
                                let mean = lowMean
                                let cvEst = estimateCv mean 
                                let standardDev = cvEst * mean 
                                d 
                                |> Seq.map (fun x -> if nan.Equals x then sampleSafe mean standardDev else x)
                            | None -> d

                let imputeUsing (x:seq<float>) = 
                    initImputeUsing sampleFunction (log2 >> log2MeanToCv) (quantileForLowMean) x

                /// Computes means and variances of complete replicates.
                let imputed = 
                    // Step 1: Set genes missing when they are not comparable
                    // because all sample groups have 1 or zero observations. 
                    grouped
                    |> Frame.getNumericCols
                    |> Series.map (fun geneID s -> 
                        let geneDataByFactorLevels = 
                            s 
                            |> Series.applyLevel fst (Series.values >> Array.ofSeq)
                        let geneData = 
                            geneDataByFactorLevels
                            |> Series.observations
                        let present = 
                            geneData 
                            |> Seq.map (fun (sampleID,v) -> 
                                    v.Length > 1
                                )
                            |> Seq.contains true 
                        if present then s else s |> Series.mapValues (fun s -> nan)
                        )
                    |> Frame.ofColumns
                    // Step 2: 
                    // x present 0 not present *nan*
                    // [x; x; x] -> [x; x; x]
                    // [x; x; 0] -> [x; x; x] by sampling, mean from sample, var by est
                    // Will be biased towards the remaining point.
                    // [x; 0; 0] -> [x; x; x] by sampling, mean from sample, var by est
                    //
                    // [0; 0; 0] -> 
                        //depending on the value of quantileForLowMean :
                        //Some x -> [x; x; x] 
                        //None   -> [0; 0; 0] 
                    |> Map.mapColWiseGroupedByFactorLevels imputeUsing

                // meanToCvEstChart |> Chart.show
                // quantilesToMeansChart |> Chart.show
                imputed

open Annotate 
open Argu
open SamPlotting
open Plotting
open MetaxaTemp
module Arguments = 
    type WorkflowArgs =
        | [<Unique>] [<Mandatory>] [<AltCommandLine("-i")>] InputPath of path:string
        //| [<Unique>] [<Mandatory>] [<AltCommandLine("-m")>] MappingFilePath of path:string
        | [<Unique>] [<Mandatory>] [<AltCommandLine("-id")>] RunID of path:string
        // | [<Unique>] [<Mandatory>] [<AltCommandLine("-o")>] OutputPath of path:string
        | [<Unique>]  [<AltCommandLine("-c")>] IdentifierColumnHeader of string
        | [<Unique>]  [<AltCommandLine("-fdr")>] FDR of string
        | [<Unique>]  [<AltCommandLine("-norm")>] PerformNormalization 
        // | [<Unique>] [<Mandatory>] [<AltCommandLine("-t")>] ColumnSeparator of string
        // |            [<Mandatory>] [<AltCommandLine("-a")>] Annotation of string
        // | [<Unique>]               [<AltCommandLine("-x")>] AnnotationSeparator of string
        // | [<Unique>]               [<AltCommandLine("-y")>] IdentifierSeparator of string
        interface IArgParserTemplate with
            member this.Usage =
                match this with
                | InputPath x                   -> "relative input file path in /arc/"
                | RunID x                       -> "ID for the run"
                //| MappingFilePath x             -> "relative path to the mapping file path in /arc/"
                // | OutputPath x                  -> "relative output file path in /arc/runs/"
                | IdentifierColumnHeader x      -> "column header of identifier column file to annotate"
                | FDR x                         -> "FDR"
                | PerformNormalization          -> "Set this flag if normalization should be performed"
                // | ColumnSeparator x             -> "column separator (tab , ; ...)"
                // | Annotation x                  -> "annotationHeader"
                // | AnnotationSeparator x         -> "multiple annotation separator"
                // | IdentifierSeparator x         -> "multiple identifier separator"
    
    let annotationArgParser = ArgumentParser.Create<WorkflowArgs>(programName = "annotateChlamy.fsx")
    
    let annotationR = 
        fsi.CommandLineArgs
        |> Array.skip 1 
        |> fun args -> annotationArgParser.ParseCommandLine(args) 
   
    // arguments are converted
    let inputPath                    = annotationR.GetResult(InputPath)//@"/arc/assays/" + annotationR.GetResult(InputPath)
    let runID                        = annotationR.GetResult(RunID)
    //let mappingFilePath              = annotationR.GetResult(MappingFilePath)
    // let outputFilePath               = annotationR.GetResult(OutputPath)
    let columnHeader                 = 
        match annotationR.TryGetResult(IdentifierColumnHeader) with 
        | None -> "Protein"
        | Some x -> x
    let fdr                 = 
        match annotationR.TryGetResult(FDR) with 
        | None -> 0.01
        | Some x -> x |> float
    let performNorm = 
        match annotationR.TryGetResult (PerformNormalization) with 
        | None -> false
        | Some x -> true
    // let separator                    = getSeparator (annotationR.GetResult(ColumnSeparator))
    // let mappings                     = annotationR.GetResults(Annotation) |> Array.ofList
    // let multipleIdentifierSeparator  = getSeparator (annotationR.GetResult(IdentifierSeparator,";"))
    // let multipleAnnotationSeparator  = getSeparator (annotationR.GetResult(AnnotationSeparator,"|"))
    
    // if separator = "," then 
    //     failwithf "\n\nWARNING: Annotation descriptions may contain comma and therefore corrupts column separation!\n"


////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////
////////////////////////////////////////
// Init OutDir 
////////////////////////////////////////
////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
let runID = Arguments.runID
System.IO.Directory.CreateDirectory runID
|> ignore
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////
////////////////////////////////////////
// Init MetaData lookup functions
////////////////////////////////////////
////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

let qi = 
    arcIO.NET.Investigation.fromArcFolder "./arc"
    |> QueryModel.QInvestigation.fromInvestigation

// qi.ProtocolNames

let allSamples =         
    qi.LastNodes 
    |> Array.ofSeq

// allSamples
// |> Array.map (fun x -> x.Name)

let getBioRep (fN:QNode) = 
    match qi.ValuesOf(fN,ProtocolName = "Growth").WithName("biological replicate").Values.Head with
    | QueryModel.ISAValue.Parameter x -> x.Value.Value.AsString 
    | _ -> failwith "no biorep please add"

// allSamples
// |> Array.map getBioRep

let getBioRepGroup (fN:QNode) = 
    match qi.ValuesOf(fN,ProtocolName = "Growth").WithName("biological replicate group").Values.Head with
    | QueryModel.ISAValue.Characteristic x -> x.Value.Value.AsString 
    | _ -> failwith "no biorep please add"

// allSamples
// |> Array.map getBioRepGroup

let qNodesOfInterest = 
    allSamples

let sampleNameToNode =
    qNodesOfInterest
    |> Array.map (fun x -> x.Name, x)
    |> Map.ofArray

let qSampleNameToIntensityColumn fN = 
    fN + " Intensity" 

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////
////////////////////////////////////////
// Execute Annotation
////////////////////////////////////////
////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
let mappingFrame = Mapping.readMapping "./arc/workflows/EvalTurboID/scripts/chlamy_jgi55.txt" "\t" "Identifier"
mappingFrame.Print()
let truncID (id : string) = (id.Split([|".p";".t"|],System.StringSplitOptions.None)).[0]
let outPathA = runID + "/AnnotatedResult.tsv"

Data.annotateAndWriteData 
    mappingFrame 
    Arguments.inputPath 
    "\t" 
    Arguments.columnHeader 
    truncID 
    [|"Synonym";"MapMan"; "MapManDescription"; "GO"; "GODescription"; "Localization";"Synonym_Phytozome_61";"GeneName_Phytozome_61"; "Description_Phytozome_61"|]
    ";"
    "|"
    outPathA
printfn "Finished Annotation" 

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////
////////////////////////////////////////
// Prepare Analysis
////////////////////////////////////////
////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
let doAnalysis (allDataForAnalysis:Frame<string,string>) dataForAnalysis (allControlSamples:QNode []) baitGroup (allBaitSamples:QNode []) = //controlsToUse normGlobal normByBait batch conditionGenotype control1 control2 inP proteinID inAssay =    
    printfn "Start Analysis: %s vs Control" baitGroup
    let allSamples = [|allBaitSamples;allControlSamples|] |> Array.concat

    let allColsOfIBait = 
        allBaitSamples
        |> Array.map (fun x -> qSampleNameToIntensityColumn x.Name)     
    
    let allColsOfIControl = 
        allControlSamples
        |> Array.map (fun x -> qSampleNameToIntensityColumn x.Name)     
    
    let quantValues :Frame<string,string> = 
        dataForAnalysis
        |> Frame.sliceCols ([|allColsOfIBait;allColsOfIControl|] |> Array.concat)

    /////////////////////////////////////////////////////
    // Normalized Data
    /////////////////////////////////////////////////////
    printfn "Normalize Data"
    /// 
    
    let normalizedAndTransposed, abundanceChart, corrFChart, corrAbundanceChart = 
        // if normGlobal then 
            let fst,ai,cor,aCorr = 
                quantValues
                |> Frame.sliceCols allColsOfIBait
                |> Frame.transpose
                |> Global.MedianOfRatios.medianOfRatiosWide
            let snd,aiCon,corCon,aCorrCon = 
                quantValues
                |> Frame.sliceCols allColsOfIControl
                |> Frame.transpose
                |> Global.MedianOfRatios.medianOfRatiosWide
            Frame.merge fst snd, 
            Chart.combine
                [
                    ai
                    |> Chart.withTraceInfo "Bait"
                    aiCon
                    |> Chart.withTraceInfo "Control"
                ], 
            Chart.combine
                [
                    cor
                    |> Chart.withTraceInfo "Bait"
                    corCon
                    |> Chart.withTraceInfo "Control"
                ],
            Chart.combine
                [
                    aCorr
                    |> Chart.withTraceInfo "Bait"
                    aCorrCon
                    |> Chart.withTraceInfo "Control"
                ]
        // else
        //     quantValuesOfInterest
        //     |> Frame.transpose
    
    /////////////////////////////////////////////////////
    // Imputation
    /////////////////////////////////////////////////////
    printfn "Impute Data"
    let descByISa :Frame<string,string> = 
        [
            "BioRepGroup",
            allSamples
            |> Seq.map (fun x -> qSampleNameToIntensityColumn x.Name, getBioRepGroup x)
            |> Series.ofObservations
            "biological replicate",
            allSamples
            |> Seq.map (fun x -> qSampleNameToIntensityColumn x.Name, getBioRep x |> string)
            |> Series.ofObservations
        ]
        |> Frame.ofColumns

    // let fittingSpline lambda (data:seq<(float*float)>) =
    //     let x,y = data |> Array.ofSeq |> Array.unzip
    //     let xV,yV = x |> vector, y |> vector
    //     let basisP = // xV |> Array.ofSeq
    //         [|
    //             for i = 0 to xV.Length-1 do 
    //                 if i = 0 then 
    //                     xV.[i]
    //                 elif i % 50 = 0 then xV.[i]
    //                 elif i = xV.Length-1 then xV.[xV.Length-1]
    //         |]

    //     FSharp.Stats.Fitting.Spline.smoothingSpline (data |> Array.ofSeq) basisP lambda

    let fitCV (data:seq<(float*float)>)  =
        // fittingSpline 1000000000000. data
        let m = 
            data 
            |> Seq.map snd 
            |> Seq.median
        fun x -> m

    let sortingInner = 
        (fun (s:ObjectSeries<string>) -> 
            let repIdx = s.GetAs<int>("biological replicate") 
            repIdx 
        )

    let sortingOuter = 
        (fun (s:ObjectSeries<string>) -> 
            let g = s.GetAs<string>("BioRepGroup")
            if g = baitGroup then 0 else 1
            // else failwith "Genotype not supported by design"        
        )

    // First we create a new !sorted! description based on a grouping created specifing the columns in the original description which are used to create a grouping. 
    let descriptionSorted =  Group.groupAndSortByFactorLevels sortingInner sortingOuter ["BioRepGroup";] (String.concat "_") descByISa

    // Then, we sort and group the data based on the new description
    let dataSorted = Group.groupRowsByGroupedFactors descriptionSorted normalizedAndTransposed

    // Then, we perform an imputation by simple gaussian sampling, if no value was recorded we sample out of the lowest 0.1 quantile of the data set.
    let imputeSamplingCondition = 
        Imputation.SimpleSampling.computeColumnWiseUsing Distributions.Continuous.Normal.Sample 3 (Some 0.01) (fitCV) 
            (dataSorted |> Frame.filterRows (fun k s -> allColsOfIBait |> Array.contains (snd k))) 
        |> Frame.mapValues log2

    let imputeSamplingControl = 
        Imputation.SimpleSampling.computeColumnWiseUsing Distributions.Continuous.Normal.Sample 3 (Some 0.01) (fitCV) 
            (dataSorted |> Frame.filterRows (fun k s -> allColsOfIControl |> Array.contains (snd k))) 
        |> Frame.mapValues log2

    let imputeSampling = 
        Frame.merge imputeSamplingCondition imputeSamplingControl

    let imputeSamplingT = 
        imputeSampling.Transpose()

    let preImpChart = 
        let ck = 
            quantValues.ColumnKeys
            |> Array.ofSeq
        quantValues
        |> Frame.mapValues log2
        |> Frame.toJaggedArray
        |> Matrix.ofJaggedArray
        |> Matrix.mapiCols (fun i c -> 
            Chart.Histogram c
            |> Chart.withTraceInfo ck.[i]
            )
        |> Chart.combine
        |> Chart.withXAxisStyle "log2 Abundance"
    let afterImpChart = 
        let ck = 
            imputeSamplingT.ColumnKeys
            |> Array.ofSeq
        imputeSamplingT
        |> Frame.toJaggedArray
        |> Matrix.ofJaggedArray
           |> Matrix.mapiCols (fun i c -> 
            Chart.Histogram c
            |> Chart.withTraceInfo ck.[i]
            )
        |> Chart.combine
        |> Chart.withXAxisStyle "log2 Abundance"
    /////////////////////////////////////////////////////
    // Norm By Quantile Normalization
    /////////////////////////////////////////////////////
    printfn "perform Quantile Normalization:%b" Arguments.performNorm
    
    let normFrame =
        if Arguments.performNorm then  
            let normM = 
                imputeSamplingT
                |> Frame.toJaggedArray
                |> Matrix.ofJaggedArray
                |> FSharp.Stats.Signal.Normalization.quantile
                |> Matrix.toArray2D
            normM
            |> Frame.ofArray2D
            |> Frame.indexRowsWith imputeSamplingT.RowKeys
            |> Frame.indexColsWith imputeSamplingT.ColumnKeys
        else 
            imputeSamplingT
    let preNormChart = 
        let ck = 
            normFrame.ColumnKeys
            |> Array.ofSeq
        imputeSamplingT
        |> Frame.toJaggedArray
        |> Matrix.ofJaggedArray
        |> Matrix.mapiCols (fun i c -> 
            Chart.Histogram c
            |> Chart.withTraceInfo ck.[i]
            )
        |> Chart.combine
        |> Chart.withXAxisStyle "log2 Abundance"
    let afterNormChart = 
        let ck = 
            normFrame.ColumnKeys
            |> Array.ofSeq
        normFrame
        |> Frame.toJaggedArray
        |> Matrix.ofJaggedArray
           |> Matrix.mapiCols (fun i c -> 
            Chart.Histogram c
            |> Chart.withTraceInfo ck.[i]
            )
        |> Chart.combine
        |> Chart.withXAxisStyle "log2 Abundance"

    /////////////////////////////////////////////////////
    // Testing
    /////////////////////////////////////////////////////
    printfn "Perform Testing"
    let (preData1,preData2) :float[][] * float [][]=  
        normFrame
        |> Frame.mapRows (fun k s -> 
            allColsOfIBait 
            |> Array.map (fun c -> s.GetAs<float>(c)),
            [|
            allColsOfIControl 
            |> Array.map (fun c -> s.GetAs<float>(c))
            |]
            |> Array.concat
            )
        |> Series.values
        |> Array.ofSeq
        |> Array.unzip

    let rowheader :string[] = normFrame.RowKeys |> Seq.toArray
    let data1 = Array.zip rowheader  preData1//preData1//
    let data2 = Array.zip rowheader  preData2//preData2//

    let res = FSharp.Stats.Testing.SAM.twoClassUnpaired 100 Arguments.fdr  data2 data1  (System.Random(1234))  
    /////////////////////////////////////////////////////
    // Summary Statistics
    /////////////////////////////////////////////////////
    printfn "Compute Summary Statistics"
    let missingValuesCondition =
        quantValues
        |> Frame.sliceCols allColsOfIBait
        |> Frame.mapRows (fun k s -> 
            allColsOfIBait
            |> Array.map (fun x -> 
                s.GetAs<float>(x,nan)
            )
            |> Array.filter (fun x -> nan.Equals x)
            |> Array.length
            )

    let missingValuesControl =
        quantValues
        |> Frame.sliceCols allColsOfIControl
        |> Frame.mapRows (fun k s -> 
            allColsOfIControl
            |> Array.map (fun x -> 
                s.GetAs<float>(x,nan)
            )
            |> Array.filter (fun x -> nan.Equals x)
            |> Array.length
            )

    let fc = 
        normFrame
        |> Frame.mapRows (fun k s -> 
            let m1 = allColsOfIBait |> Seq.map (fun x -> s.GetAs<float>(x,nan)) |> Seq.mean
            let m2 = allColsOfIControl   |> Seq.map (fun x -> s.GetAs<float>(x,nan)) |> Seq.mean
            m1-m2
        )

    let maxMeanAbundance = 
        normFrame
        |> Frame.mapRows (fun k s -> 
            let m1 = allColsOfIBait |> Seq.map (fun x -> s.GetAs<float>(x,nan)) |> Seq.mean
            let m2 = allColsOfIControl   |> Seq.map (fun x -> s.GetAs<float>(x,nan)) |> Seq.mean
            if m1 > m2 then m1 else m2
        )

    let meanAbundanceBait = 
        normFrame
        |> Frame.mapRows (fun k s -> 
            let m1 = allColsOfIBait |> Seq.map (fun x -> s.GetAs<float>(x,nan)) |> Seq.mean
            m1
        )

    let meanAbundanceControl = 
        normFrame
        |> Frame.mapRows (fun k s ->           
            let m2 = allColsOfIControl   |> Seq.map (fun x -> s.GetAs<float>(x,nan)) |> Seq.mean
            m2
        )    
    let stdev = 
        normFrame
        |> Frame.sliceCols allColsOfIBait 
        |> Frame.transpose
        |> Stats.stdDev
    let stdevControl = 
        normFrame
        |> Frame.sliceCols allColsOfIControl 
        |> Frame.transpose
        |> Stats.stdDev
    let qValue = 
        [|res.PosSigBioitem;res.NegSigBioitem;res.NonSigBioitem|]
        |> Array.concat
        |> Array.map (fun x -> x.ID,x.QValue)
        |> Series.ofObservations

    let IsCandidate = 
        [|res.PosSigBioitem;res.NegSigBioitem;res.NonSigBioitem|]
        |> Array.concat
        |> Array.map (fun x -> x.ID,x.QValue < Arguments.fdr )
        |> Series.ofObservations
    
    let finPlot = 
        normFrame
        |> Frame.addCol "MeanR" fc
        |> Frame.addCol "MeanAbundance" maxMeanAbundance
        |> Frame.addCol "StDevR" stdev
        |> Frame.addCol "qValue" qValue
        |> Frame.addCol "missingValuesCondition" missingValuesCondition
        |> Frame.addCol "missingValuesControl" missingValuesControl
        |> Frame.addCol "IsCandidate" IsCandidate
        |> Frame.addCol "ProtName" 
            (allDataForAnalysis 
            |> Frame.mapRows (fun k s -> 
                [|s.GetAs<string>("Synonym_Phytozome_61","");s.GetAs<string>("GeneName_Phytozome_61","");k|] 
                |> String.concat ";"
                |> fun x -> x.Replace('|',';').Split(';')
                |> Array.distinct
                |> Array.filter (fun x -> x <> "")
                |> String.concat ";"
                )   
            )//|> Array.distinct |> String.concat ";"))//(allDataForAnalysis.GetColumn<string>("Synonym"))   
    
    let fin = 
        normFrame
        |> Frame.addCol "log2FC_Bait_Minus_Control" fc
        |> Frame.addCol "MaxMeanLog2Abundance" maxMeanAbundance
        |> Frame.addCol "MeanLog2Abundance_Bait" meanAbundanceBait
        |> Frame.addCol "StDev_Bait" stdev
        |> Frame.addCol "MeanLog2Abundance_Control" meanAbundanceControl
        |> Frame.addCol "StDev_Control" stdevControl
        |> Frame.addCol "qValue" qValue
        |> Frame.addCol "MissingValues_Bait" missingValuesCondition
        |> Frame.addCol "MissingValues_Control" missingValuesControl
        |> Frame.addCol "IsSignificant" IsCandidate
        
    let toSave = Frame.join JoinKind.Outer fin (Frame.sliceCols ["Synonym"; "MapMan";"MapManDescription"; "GO"; "GODescription";"Localization";"Synonym_Phytozome_61";"GeneName_Phytozome_61"; "Description_Phytozome_61"] allDataForAnalysis)    
    
    /////////////////////////////////////////////////////
    // Save Results
    /////////////////////////////////////////////////////
    let outP = System.IO.Path.Combine([|runID;baitGroup|])
    System.IO.Directory.CreateDirectory outP
    |> ignore

    abundanceChart
    |> Chart.withTemplate ChartTemplates.lightMirrored
    |> Chart.saveHtml (outP + @"/1_Norm_PreNormGlobalAbundanceChart.html")
    corrFChart
    |> Chart.withTemplate ChartTemplates.lightMirrored
    |> Chart.saveHtml (outP + @"/2_Norm_CorrectionfactorChart.html")
    corrAbundanceChart
    |> Chart.withTemplate ChartTemplates.lightMirrored
    |> Chart.saveHtml (outP + @"/3_Norm_AfterNormGlobalAbundanceChart.html")
    preImpChart
    |> Chart.withTemplate ChartTemplates.lightMirrored
    |> Chart.saveHtml (outP + @"/4_Imputation_preImpChart.html")
    afterImpChart
    |> Chart.withTemplate ChartTemplates.lightMirrored
    |> Chart.saveHtml (outP + @"/5_Imputation_afterImpChart.html")
    preNormChart
    |> Chart.withTemplate ChartTemplates.lightMirrored
    |> Chart.saveHtml (outP + @"/6_QuantilNorm_PreNormChart.html")
    afterNormChart
    |> Chart.withTemplate ChartTemplates.lightMirrored
    |> Chart.saveHtml (outP + @"/7_QuantilNorm_AfterNormChart.html")
    createSAMChart res
    |> Chart.withTemplate ChartTemplates.lightMirrored
    |> Chart.saveHtml (outP + @"/8_Testing_SAM.html")
    plotHisto finPlot
    |> Chart.withTemplate ChartTemplates.lightMirrored
    |> Chart.saveHtml (outP + @"/9_Testing_DifferencesHistogram.html")
    plotMA [] finPlot // ["CDJ5";"LON";"PSBQLD";"PGRL1"]
    |> Chart.withTemplate ChartTemplates.lightMirrored
    |> Chart.saveHtml (outP + @"/10_Testing_MAPlot.html")
    plotVulcano [] finPlot // ["CDJ5";"LON";"PSBQLD";"PGRL1"]
    |> Chart.withTemplate ChartTemplates.lightMirrored
    |> Chart.saveHtml (outP + @"/11_Testing_VulcanoPlot.html")
    toSave.SaveCsv(outP + @"/Analysis_complete.tsv",includeRowKeys=true,keyNames=["Protein"],separator='\t')
    System.IO.File.WriteAllLines(outP + @"/params.txt",fsi.CommandLineArgs)
    printfn "Finished Analysis! results can be found at:%s" outP 

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////
////////////////////////////////////////
// Execute Analysis
////////////////////////////////////////
///////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
let allDataForAnalysis :Frame<string,string> = 
    Frame.ReadCsv(path = outPathA, hasHeaders=true, separators="\t")
    |> Frame.indexRows Arguments.columnHeader 

let quantDataForAnalysis :Frame<string,string> = 
    let schema = 
        Frame.ReadCsv(path = outPathA, hasHeaders=true, separators="\t").ColumnKeys
        |> Seq.filter (fun x -> x.Contains "Intensity")
        |> Seq.map (fun x -> sprintf "%s=float" x)
        |> String.concat ","
    Frame.ReadCsv(path = outPathA, hasHeaders=true, separators="\t",schema= schema)
    |> Frame.indexRows Arguments.columnHeader 
    |> Frame.mapValues (fun x -> if x > 0. then x else nan)

let samplesGroupedByBioRepGroup = 
    allSamples
    |> Array.map (fun x -> 
        x, getBioRepGroup x)
    |> Array.groupBy snd
    |> Array.map (fun (x,y) -> 
        {|
        BaitGroup = x 
        QNodes = 
            y 
            |> Array.map fst 
        |}
        )
    
let allBaitGroups = 
    samplesGroupedByBioRepGroup
    |> Array.filter (fun x -> x.BaitGroup <> "control")

let allControlSamples = 
    samplesGroupedByBioRepGroup
    |> Array.filter (fun x ->  x.BaitGroup = "control")
    |> Array.map (fun x -> x.QNodes) 
    |> Array.concat

allBaitGroups
|> Array.iter (fun g -> doAnalysis allDataForAnalysis quantDataForAnalysis allControlSamples g.BaitGroup g.QNodes)    

printfn "Done!" 





