#r "nuget: arcIO.NET, 0.0.1"
#r "nuget: ISADotNet.QueryModel, 0.5.4"
#r "nuget: BioFSharp.IO, 2.0.0-preview.3"
#r "nuget: FSharp.Stats, 0.4.9"
#r "nuget: Plotly.NET, 3.0.1"
// #r "nuget: Deedle, 3.0.0-beta.1"
#r "nuget: Deedle"
#r "nuget: DeedleExtensions, 0.2.0"
open FSharp.Stats
open Plotly.NET
open Deedle
open BioFSharp.IO
open DeedleExtensions
open ISADotNet

type UseControl = 
    | C4
    | C13
    | Both


module Global = 
    /// As used by Deseq2, see: https://github.com/hbctraining/DGE_workshop/blob/master/lessons/02_DGE_count_normalization.md 
    module MedianOfRatios = 

        /// expects data frame with shape frame<sampleIds,GeneIds>
        let medianOfRatiosWide (data:Frame<string,string>) =
            //data frame with shape frame<sampleIds,GeneIds>
            let ratios = 
                data
                |> Frame.dropSparseCols
                |> Frame.getNumericCols
                |> Series.map (fun geneID s -> 
                    let geometricMeanAcrossAllSamples = 
                        s.Values 
                        |> FSharp.Stats.Seq.meanGeometric
                    let ratioPerSample = 
                        s
                        |> Series.map (fun x sample ->  sample / geometricMeanAcrossAllSamples)
                    ratioPerSample
                    )
                |> Frame.ofColumns
            // Series of shape: Series<sampleId,correctionFactor:float>
            let correctionFactorPerSample = 
                ratios
                |> Frame.transpose
                |> Stats.median
            // Chart.Column(correctionFactorPerSample |> Series.observations)
            // |> Chart.withXAxisStyle "SampleId"
            // |> Chart.withXAxisStyle "Correction Factor."
            //  |> Chart.show
            // data frame with frame<sampleIds,GeneIds>
            // contains gene data divided by the sample wise estimated correction factor
            let correctedData = 
                data
                |> Frame.getNumericCols 
                |> Series.map (fun geneID s -> 
                    s / correctionFactorPerSample
                    )
                |> Frame.ofColumns
            correctedData

        /// expects data frame with shape frame<GeneIds,sampleIds>
        let medianOfRatios (data:Frame<string,string>) =
            data
            |> Frame.transpose
            |> medianOfRatiosWide
            |> Frame.transpose

let doAnalysis controlsToUse normGlobal normByBait batch conditionGenotype control1 control2 inP proteinID inAssay = 
    ///////////////////////////////////////////////////////
    /// params
    ///////////////////////////////////////////////////////
    let batch = batch//2
    let conditionGenotype = conditionGenotype//"UVM_CDJ5_WT_TurboID"
    let control1 = control1//"UVM_Cher4"
    let control2 = control2//"UVM_Cher13"
    let inP  = inP//__SOURCE_DIRECTORY__ + @"\..\..\runs\4_CDJ5_MSFraggerCombined_annotated\MSFraggerCombined_annotated.tsv" 
    let proteinID = proteinID//"Protein" 
    let inAssay = inAssay//"CDJ5Proteomics"

    ///////////////////////////////////////////////////////
    /// Isa
    ///////////////////////////////////////////////////////


    let arcPath = __SOURCE_DIRECTORY__ + @"\..\..\"

    let p,a = arcIO.Net.Assay.readByName arcPath inAssay

    let qa = QueryModel.QStudy.fromAssay a

    let allSamples = 
        qa.LastNodes()
        |> Set.ofSeq

    let getBioRep fN = 
        qa.ValuesOf(fN,ProtocolName = "Growth").WithName("biological replicate").First.ValueWithUnitText
        |> int       

    let getGenotype fN = 
        qa.ValuesOf(fN,ProtocolName = "GenotypeLib").WithName("Genotype").First.ValueWithUnitText
        |> string   

    let getBatch fN = 
        qa.ValuesOf(fN,ProtocolName = "Growth").WithName("Batch").First.ValueWithUnitText
        |> int    

    let getTurboIDIdentifier fN = 
        qa.ValuesOf(fN,ProtocolName = "GenotypeLib").WithName("gene identifier").First.ValueWithUnitText
        |> string        

    ///////////////////////////////////////////////////////
    /// Raw Data
    ///////////////////////////////////////////////////////
    let schema = 
        allSamples
        |> Seq.map (fun x -> sprintf "%s = float" x)
        |> String.concat ","

    ///
    let all :Frame<string,string> = 
        Frame.ReadCsv(path = inP, hasHeaders=true, separators="\t",schema= schema)
        |> Frame.indexRows proteinID 
        |> Frame.mapValues (fun x -> if x > 0. then x else nan)
        
    let quantValues :Frame<string,string> = 
        all
        |> Frame.sliceCols allSamples


    let normalized = 
        if normGlobal then 
            quantValues
            |> Global.MedianOfRatios.medianOfRatios
        else
            quantValues

    let allColumnsCondition = 
        normalized.ColumnKeys 
        |> Seq.filter (fun x -> getGenotype x = conditionGenotype) 
        |> Array.ofSeq

    let allColumnsCher4Control = 
        normalized.ColumnKeys 
        |> Seq.filter (fun x -> getGenotype x = control1 && getBatch x = batch) 
        |> Array.ofSeq
        
    let allColumnsCher13Control = 
        normalized.ColumnKeys 
        |> Seq.filter (fun x -> getGenotype x = control2 && getBatch x = batch) 
        |> Array.ofSeq


    let calcRep norm (f:Frame<string,string>) (rep:int) = 
        let condC = allColumnsCondition |> Array.find (fun x -> getBioRep x = rep)
        let cher4Control = allColumnsCher4Control |> Array.find (fun x -> getBioRep x = rep)
        let cher13Control = allColumnsCher13Control |> Array.find (fun x -> getBioRep x = rep)

        let CGE = f.GetColumn<float>(condC)    
        let _13 = f.GetColumn<float>(cher13Control)  
        let _4  = f.GetColumn<float>(cher4Control)  

        let cgeBaitFactor = CGE |> Series.get (getTurboIDIdentifier condC)
        let _13BaitFactor = _13 |> Series.get (getTurboIDIdentifier cher13Control)
        let _4BaitFactor = _4 |> Series.get (getTurboIDIdentifier cher4Control)

        let NormAndLog2_Condition = 
            Series.zip CGE _13
            |> Series.mapValues (fun (prot_cge,prot_13) -> 
                    match prot_cge.HasValue, prot_13.HasValue with  
                    | true,true -> 
                        let scaled1 = 
                            if norm then 
                                prot_cge.Value / cgeBaitFactor
                            else
                                prot_cge.Value 
                        log10 scaled1
                    | _ -> nan
                )


        let NormAndLog2_Cher13_1 = 
            Series.zip CGE _13
            |> Series.mapValues (fun (prot_cge,prot_13) -> 
                    match prot_cge.HasValue, prot_13.HasValue with  
                    | true,true -> 
                        let scaled =
                            if norm then 
                                prot_13.Value / _13BaitFactor
                            else
                                prot_13.Value 
                        log10 scaled
                    | _ -> nan
                )

        
        let NormAndLog2_Cher4_2 = 
            Series.zip CGE _4
            |> Series.mapValues (fun (prot_cge,prot_4) -> 
                    match prot_cge.HasValue, prot_4.HasValue with 
                    | true, true -> 
                        let scaled =
                            if norm then 
                                prot_4.Value / _4BaitFactor
                            else prot_4.Value
                        log10 scaled
                    | _ -> nan
                )            
        [
            (rep |> string) + "_NormAndLog10_Condition" ,NormAndLog2_Condition;
            (rep |> string) + "_NormAndLog10_Cher13_1"  ,NormAndLog2_Cher13_1;
            (rep |> string) + "_NormAndLog10_Cher4_2"   ,NormAndLog2_Cher4_2;
            (rep |> string) + "_abundance_cge",CGE;
            (rep |> string) + "_abundance_13",_4;
            (rep |> string) + "_abundance_4",_4
        ]
        |> Frame.ofColumns


    let calculateStats controlsToUse norm (f:Frame<string,string>) :Frame<string,string>=
        
        // Todo: get replicates to iter from sample description
        let r3s = 
            [calcRep norm f 1;calcRep norm f 2;calcRep norm f 3;calcRep norm f 4]
            |> Frame.mergeAll 

        let aCols = r3s.ColumnKeys |> Seq.filter (fun x -> x.Contains "abundance")
        let NormAndLog10_ConditionCols = r3s.ColumnKeys |> Seq.filter (fun x -> x.Contains "NormAndLog10_Condition") |> Array.ofSeq
        let NormAndLog10_Cher13_1Cols = r3s.ColumnKeys |> Seq.filter (fun x -> x.Contains "NormAndLog10_Cher13_1") |> Array.ofSeq
        let NormAndLog10_Cher4_2Cols = r3s.ColumnKeys |> Seq.filter (fun x -> x.Contains "NormAndLog10_Cher4_2") |> Array.ofSeq
            
        let meanRatioNoSparse =  
            r3s
            |> Frame.dropSparseRows
            |> Frame.mapRows (fun k s -> 
                    NormAndLog10_ConditionCols 
                    |> Array.map (fun ck -> 
                        let valCondition = s.GetAs<float>(ck)
                        let partnerColCher13 = Array.find (fun (x:string) -> ck.Split('_').[0] = x.Split('_').[0]) NormAndLog10_Cher13_1Cols
                        let valCher13 = s.GetAs<float>(partnerColCher13)
                        let partnerColCher4 = Array.find (fun (x:string) -> ck.Split('_').[0] = x.Split('_').[0]) NormAndLog10_Cher4_2Cols
                        let valCher4 = s.GetAs<float>(partnerColCher4)
                        [|
                            valCondition-valCher13;
                            valCondition-valCher4
                        |]
                        )
                    |> Array.concat
                    |> Seq.mean
                )
            |> Series.values
            |> Array.ofSeq

        let cutOff = 
            let borders =
                meanRatioNoSparse
                |> Array.ofSeq
                |> FSharp.Stats.Testing.Outliers.tukey 1.5
            borders.Upper
        let highestRatio :float = meanRatioNoSparse |> Seq.max
        let lowestRatio  :float = meanRatioNoSparse |> Seq.min
        let stats = 
            r3s
            |> Frame.mapRows (fun k s -> 
                    let asOfProt = aCols |> Seq.map (fun ck -> s.GetAs<float>(ck,nan)) |> Seq.filter (fun x -> nan.Equals x |> not)
                    let cgeAbundances = aCols |> Seq.filter (fun x -> x.Contains "cge") |> Seq.map (fun ck -> s.GetAs<float>(ck,nan)) |> Seq.filter (fun x -> nan.Equals x |> not)                
                    let mCherryAbundances = aCols |> Seq.filter (fun x -> x.Contains "cge" |> not) |> Seq.map (fun ck -> s.GetAs<float>(ck,nan)) |> Seq.filter (fun x -> nan.Equals x |> not)                
                    let meanAbundance = asOfProt |> Seq.map log10 |> Seq.mean
                    let rsCher13 = 
                        NormAndLog10_ConditionCols 
                        |> Array.map (fun ck -> 
                            let valCondition = s.GetAs<float>(ck,nan)
                            let partnerColCher13 = Array.find (fun (x:string) -> ck.Split('_').[0] = x.Split('_').[0]) NormAndLog10_Cher13_1Cols
                            let valCher13 = s.GetAs<float>(partnerColCher13,nan)
                            valCondition-valCher13
                        )
                        |> Array.filter (fun x -> nan.Equals x |> not)
                    let rsCher4 = 
                        NormAndLog10_ConditionCols 
                        |> Array.map (fun ck -> 
                            let valCondition = s.GetAs<float>(ck,nan)
                            let partnerColCher4 = Array.find (fun (x:string) -> ck.Split('_').[0] = x.Split('_').[0]) NormAndLog10_Cher4_2Cols
                            let valCher4 = s.GetAs<float>(partnerColCher4,nan)
                            valCondition-valCher4
                            )
                        |> Array.filter (fun x -> nan.Equals x |> not)
                    let nRatios = (rsCher4 |> Seq.length) + (rsCher13 |> Seq.length)
                    let nQuantifiedInCGE = cgeAbundances |> Seq.length
                    let nQuantifiedInMCherry = mCherryAbundances |> Seq.length
                    let meanR = 
                        if nRatios > 0 then
                            match controlsToUse with 
                            | Both -> 
                                Seq.mean (Array.concat [|rsCher13;rsCher4|])
                            | C13 ->
                                Seq.mean (Array.concat [|rsCher13;|])
                            | C4 ->
                                Seq.mean (Array.concat [|rsCher4;|])
                        elif nQuantifiedInCGE > 0 && nQuantifiedInMCherry = 0 then 
                            highestRatio 
                        elif nQuantifiedInMCherry > 0 && nQuantifiedInCGE = 0 then 
                            lowestRatio     
                        else nan           
                    let stDevR = 
                        if nRatios > 1 then
                            match controlsToUse with 
                            | Both -> 
                                Seq.mean (Array.concat [|rsCher13;rsCher4|])
                            | C13 ->
                                Seq.mean (Array.concat [|rsCher13;|])
                            | C4 ->
                                Seq.mean (Array.concat [|rsCher4;|])
                        else 
                            0.
                    let isCandidate = meanR > cutOff
                    {|
                        Index           = k
                        IsCandidate     = isCandidate
                        MeanAbundance   = meanAbundance
                        MeanR           = meanR
                        StDevR          = stDevR
                        NRatios         = nRatios
                        NQuantifiedInCGE = nQuantifiedInCGE
                    |}
                )
            |> Series.values
        stats
        |> Frame.ofRecords
        |> Frame.indexRows "Index"
    let resLog = calculateStats controlsToUse normByBait  normalized
    resLog
    |> Frame.addCol "ProtName" (all.GetColumn<string>("Synonym"))

#load "Plotting.fsx"
open Plotting

// let batch = 2
// let conditionGenotype = "UVM_CDJ5_WT_TurboID"
let control1 = "UVM_Cher4"
let control2 = "UVM_Cher13"
let inP  = __SOURCE_DIRECTORY__ + @"\..\..\runs\4_CDJ5_MSFraggerCombined_annotated\MSFraggerCombined_annotated.tsv" 
let proteinID = "Protein" 
let inAssay = "CDJ5Proteomics"


let doPlotForSpecificCondition2 controlsToUse batch globalNorm conditionGenotype = 
    let resNoNorm = doAnalysis controlsToUse globalNorm false batch conditionGenotype control1 control2 inP proteinID inAssay 

    plotRatiosVsMeanAbundanceSpec ["CDJ5";"LON";"PSBQLD";"PGRL1"] resNoNorm
    |> Chart.withTitle (sprintf "resNorm, globalnorm:%A, control:%A" globalNorm conditionGenotype)
    |> Chart.show


    // let resNorm = doAnalysis controlsToUse globalNorm true batch conditionGenotype control1 control2 inP proteinID inAssay 

    // plotRatiosVsMeanAbundanceSpec ["CDJ5";"LON";"PSBQLD";"PGRL1"] resNorm
    // |> Chart.withTitle (sprintf "resNorm, globalnorm:%A, control:%A" globalNorm conditionGenotype)
    // |> Chart.show

    [
    plotHistoFixBW 0.05 resNoNorm
    |> Chart.withTraceInfo "resNoNorm"
    // plotHistoFixBW 0.05 resNorm
    // |> Chart.withTraceInfo "resNorm"
    ]
    |> Chart.combine
    |> Chart.withTitle conditionGenotype
    |> Chart.show

// doPlotForSpecificCondition2 UseControl.Both 1 false "UVM_CDJ5_AAA_TurboID"
// doPlotForSpecificCondition2 UseControl.C13 1 false "UVM_CDJ5_AAA_TurboID"
// doPlotForSpecificCondition2 UseControl.C4 1 false "UVM_CDJ5_AAA_TurboID"
// // doPlotForSpecificCondition2 UseControl.Both 2 false "UVM_CDJ5_AAA_TurboID"
// // doPlotForSpecificCondition2 UseControl.C13 2 false "UVM_CDJ5_AAA_TurboID"
// // doPlotForSpecificCondition2 UseControl.C4 2 false "UVM_CDJ5_AAA_TurboID"


// doPlotForSpecificCondition2 UseControl.Both 1 false "UVM_CDJ5_WT_TurboID"
// doPlotForSpecificCondition2 UseControl.C13 1 false  "UVM_CDJ5_WT_TurboID"
// doPlotForSpecificCondition2 UseControl.C4 1 false   "UVM_CDJ5_WT_TurboID"
// // doPlotForSpecificCondition2 UseControl.Both 2 false "UVM_CDJ5_WT_TurboID"
// // doPlotForSpecificCondition2 UseControl.C13 2 false  "UVM_CDJ5_WT_TurboID"
// // doPlotForSpecificCondition2 UseControl.C4 2 false   "UVM_CDJ5_WT_TurboID"


doPlotForSpecificCondition2 UseControl.Both 1 true "UVM_CDJ5_WT_TurboID"
doPlotForSpecificCondition2 UseControl.C13 1 true  "UVM_CDJ5_WT_TurboID"
doPlotForSpecificCondition2 UseControl.C4 1 true   "UVM_CDJ5_WT_TurboID"
// doPlotForSpecificCondition2 UseControl.Both 2 true "UVM_CDJ5_WT_TurboID"
// doPlotForSpecificCondition2 UseControl.C13 2 true  "UVM_CDJ5_WT_TurboID"
// doPlotForSpecificCondition2 UseControl.C4 2 true   "UVM_CDJ5_WT_TurboID"
// doPlotForSpecificCondition2 1 true "UVM_CDJ5_AAA_TurboID"

// doPlotForSpecificCondition2 2 false "UVM_CDJ5_AAA_TurboID"
// doPlotForSpecificCondition2 2 true "UVM_CDJ5_AAA_TurboID"

// doPlotForSpecificCondition2 1 false "UVM"
// doPlotForSpecificCondition2 1 true "UVM"

doPlotForSpecificCondition2 UseControl.Both 1 false "UVM_CDJ5_WT_TurboID"
doPlotForSpecificCondition2 UseControl.Both 1 true "UVM_CDJ5_WT_TurboID"

doPlotForSpecificCondition2 UseControl.Both 2 false "UVM_Lon_TurboID"
doPlotForSpecificCondition2 UseControl.Both 2 true "UVM_Lon_TurboID"

doPlotForSpecificCondition2 UseControl.C13 2 false "UVM_Lon_TurboID"
doPlotForSpecificCondition2 UseControl.C13 2 true "UVM_Lon_TurboID"

doPlotForSpecificCondition2 UseControl.C4 2 false "UVM_Lon_TurboID"
doPlotForSpecificCondition2 UseControl.C4 2 true "UVM_Lon_TurboID"

doPlotForSpecificCondition2 UseControl.Both 2 false "UVM_PsbQ_TurboID"
doPlotForSpecificCondition2 UseControl.Both 2 true "UVM_PsbQ_TurboID"

doPlotForSpecificCondition2 UseControl.Both 2 false "UVM_TEF_TurboID"
doPlotForSpecificCondition2 UseControl.Both 2 true "UVM_TEF_TurboID"