#r "nuget: arcIO.NET, 0.0.1"
#r "nuget: ISADotNet.QueryModel, 0.5.4"
#r "nuget: BioFSharp.IO, 2.0.0-preview.3"
#r "nuget: FSharp.Stats, 0.4.9"
#r "nuget: Plotly.NET, 3.0.1"
// #r "nuget: Deedle, 3.0.0-beta.1"
#r "nuget: Deedle"
#r "nuget: DeedleExtensions, 0.2.0"
open FSharp.Stats
open Plotly.NET
open Deedle
open BioFSharp.IO
open DeedleExtensions


///////////////////////////////////////////////////////
/// params
///////////////////////////////////////////////////////
let batch = 1
let conditionGenotype = "UVM_CDJ5_WT_TurboID"
let control1 = "UVM_Cher4"
let control2 = "UVM_Cher13"
let inP  = __SOURCE_DIRECTORY__ + @"\..\runs\CDJ5_MSFraggerCombined\MSFraggerCombined.tsv" 
let proteinID = "Protein" 
let inAssay = "CDJ5Proteomics"

///////////////////////////////////////////////////////
/// Isa
///////////////////////////////////////////////////////
open ISADotNet

let arcPath = __SOURCE_DIRECTORY__ + @"\..\"

let p,a = arcIO.Net.Assay.readByName arcPath inAssay

let qa = QueryModel.QStudy.fromAssay a

let allSamples = 
    qa.LastNodes()
    |> Set.ofSeq

let getBioRep fN = 
    qa.ValuesOf(fN,ProtocolName = "Growth").WithName("biological replicate").First.ValueWithUnitText
    |> int       

let getGenotype fN = 
    qa.ValuesOf(fN,ProtocolName = "GenotypeLib").WithName("Genotype").First.ValueWithUnitText
    |> string   

let getBatch fN = 
    qa.ValuesOf(fN,ProtocolName = "Growth").WithName("Batch").First.ValueWithUnitText
    |> int    

let getTurboIDIdentifier fN = 
    qa.ValuesOf(fN,ProtocolName = "GenotypeLib").WithName("gene identifier").First.ValueWithUnitText
    |> string        

///////////////////////////////////////////////////////
/// Raw Data
///////////////////////////////////////////////////////
let schema = 
    allSamples
    |> Seq.map (fun x -> sprintf "%s = float" x)
    |> String.concat ","

///
let quantValues :Frame<string,string> = 
    Frame.ReadCsv(path = inP, hasHeaders=true, separators="\t",schema= schema)
    |> Frame.indexRows proteinID 
    |> Frame.mapValues (fun x -> if x > 0. then x else nan)

module Global = 
    /// As used by Deseq2, see: https://github.com/hbctraining/DGE_workshop/blob/master/lessons/02_DGE_count_normalization.md 
    module MedianOfRatios = 

        /// expects data frame with shape frame<sampleIds,GeneIds>
        let medianOfRatiosWide (data:Frame<string,string>) =
            //data frame with shape frame<sampleIds,GeneIds>
            let ratios = 
                data
                |> Frame.dropSparseCols
                |> Frame.getNumericCols
                |> Series.map (fun geneID s -> 
                    let geometricMeanAcrossAllSamples = 
                        s.Values 
                        |> FSharp.Stats.Seq.meanGeometric
                    let ratioPerSample = 
                        s
                        |> Series.map (fun x sample ->  sample / geometricMeanAcrossAllSamples)
                    ratioPerSample
                    )
                |> Frame.ofColumns
            // Series of shape: Series<sampleId,correctionFactor:float>
            let correctionFactorPerSample = 
                ratios
                |> Frame.transpose
                |> Stats.median
            Chart.Column(correctionFactorPerSample |> Series.observations)
            |> Chart.withXAxisStyle "SampleId"
            |> Chart.withXAxisStyle "Correction Factor."
             |> Chart.show
            // data frame with frame<sampleIds,GeneIds>
            // contains gene data divided by the sample wise estimated correction factor
            let correctedData = 
                data
                |> Frame.getNumericCols 
                |> Series.map (fun geneID s -> 
                    s / correctionFactorPerSample
                    )
                |> Frame.ofColumns
            correctedData

        /// expects data frame with shape frame<GeneIds,sampleIds>
        let medianOfRatios (data:Frame<string,string>) =
            data
            |> Frame.transpose
            |> medianOfRatiosWide
            |> Frame.transpose

let normalized = 
    quantValues
    |> Global.MedianOfRatios.medianOfRatios

let allColumnsCondition = 
    normalized.ColumnKeys 
    |> Seq.filter (fun x -> getGenotype x = conditionGenotype) 
    |> Array.ofSeq

let allColumnsCher4Control = 
    normalized.ColumnKeys 
    |> Seq.filter (fun x -> getGenotype x = control1 && getBatch x = batch) 
    |> Array.ofSeq
    
let allColumnsCher13Control = 
    normalized.ColumnKeys 
    |> Seq.filter (fun x -> getGenotype x = control2 && getBatch x = batch) 
    |> Array.ofSeq

let calcRep (f:Frame<string,string>) (rep:int) = 
    let condC = allColumnsCondition |> Array.find (fun x -> getBioRep x = rep)
    let cher4Control = allColumnsCher4Control |> Array.find (fun x -> getBioRep x = rep)
    let cher13Control = allColumnsCher13Control |> Array.find (fun x -> getBioRep x = rep)

    let CGE = f.GetColumn<float>(condC)    
    let _13 = f.GetColumn<float>(cher13Control)  
    let _4  = f.GetColumn<float>(cher4Control)  
    let r1_13  = (CGE |> Series.get (getTurboIDIdentifier condC)) / (_13 |> Series.get (getTurboIDIdentifier cher13Control))
    let r1_4   = (CGE |> Series.get (getTurboIDIdentifier condC)) / (_4 |> Series.get (getTurboIDIdentifier cher4Control))
    
    let r3s_13 = 
        Series.zip CGE _13
        |> Series.mapValues (fun (prot_cge,prot_13) -> 
                match prot_cge.HasValue, prot_13.HasValue with  
                | true,true -> 
                    let r2_13 = prot_cge.Value / prot_13.Value
                    r2_13 / r1_13
                | _ -> nan
            )
    
    let r3s_4 = 
        Series.zip CGE _4
        |> Series.mapValues (fun (prot_cge,prot_4) -> 
                match prot_cge.HasValue, prot_4.HasValue with 
                | true, true -> 
                    let r2_4 = prot_cge.Value / prot_4.Value
                    r2_4 / r1_4
                | _ -> nan
            )

    let diff_13 = 
        Series.zip CGE _13
        |> Series.mapValues (fun (prot_cge,prot_13) -> 
                match prot_cge.HasValue, prot_13.HasValue with  
                | true,true -> 
                    let scaled1 = prot_cge.Value /  (CGE |> Series.get (getTurboIDIdentifier condC))
                    let scaled2 = prot_13.Value /   (_13 |> Series.get (getTurboIDIdentifier cher13Control))
                    scaled1 - scaled2
                | _ -> nan
            )

    
    let diff_4 = 
        Series.zip CGE _4
        |> Series.mapValues (fun (prot_cge,prot_4) -> 
                match prot_cge.HasValue, prot_4.HasValue with 
                | true, true -> 
                    let scaled1 = prot_cge.Value /  (CGE |> Series.get (getTurboIDIdentifier condC))
                    let scaled2 = prot_4.Value /   (_4 |> Series.get (getTurboIDIdentifier cher4Control))
                    scaled1 - scaled2
                | _ -> nan
            )            
    [
        (rep |> string) + "_r3s_13",r3s_13;
        (rep |> string) + "_r3s_4",r3s_4;
        (rep |> string) + "diff_13",diff_13;
        (rep |> string) + "diff_4" ,diff_4;
        (rep |> string) + "_abundance_cge",CGE;
        (rep |> string) + "_abundance_13",_4;
        (rep |> string) + "_abundance_4",_4
    ]
    |> Frame.ofColumns




let calculateStats (f:Frame<string,string>) :Frame<string,string>=
    
    let r3s = 
        [calcRep f 1;calcRep f 2;calcRep f 3;calcRep f 4]
        |> Frame.mergeAll 
    
    let rCols = r3s.ColumnKeys |> Seq.filter (fun x -> x.Contains "r3s")
    let aCols = r3s.ColumnKeys |> Seq.filter (fun x -> x.Contains "abundance")
    let dCols = r3s.ColumnKeys |> Seq.filter (fun x -> x.Contains "diff_")

    let meanRNoSparse =  
        r3s
        |> Frame.dropSparseRows
        |> Frame.mapRows (fun k s -> 
                let rsOfProt = rCols |> Seq.map (fun ck -> s.GetAs<float>(ck))
                let meanR = Seq.mean rsOfProt
                meanR
            )
        |> Series.values
    
    let cutOff = 
        let borders =
            meanRNoSparse
            |> Array.ofSeq
            |> FSharp.Stats.Testing.Outliers.tukey 1.5
        borders.Upper

    let highestRatio :float = meanRNoSparse |> Seq.max
    let lowestRatio  :float = meanRNoSparse |> Seq.min
    let stats = 
        r3s
        // |> Frame.dropSparseRows
        |> Frame.mapRows (fun k s -> 
                let asOfProt = aCols |> Seq.map (fun ck -> s.GetAs<float>(ck,nan)) |> Seq.filter (fun x -> nan.Equals x |> not)
                let cgeAbundances = aCols |> Seq.filter (fun x -> x.Contains "cge") |> Seq.map (fun ck -> s.GetAs<float>(ck,nan)) |> Seq.filter (fun x -> nan.Equals x |> not)                
                let mCherryAbundances = aCols |> Seq.filter (fun x -> x.Contains "cge" |> not) |> Seq.map (fun ck -> s.GetAs<float>(ck,nan)) |> Seq.filter (fun x -> nan.Equals x |> not)                
                let meanAbundance = asOfProt |> Seq.mean
                let rsOfProt = rCols |> Seq.map (fun ck -> s.GetAs<float>(ck,nan)) |> Seq.filter (fun x -> nan.Equals x |> not)
                let dsOfProt = dCols |> Seq.map (fun ck -> s.GetAs<float>(ck,nan)) |> Seq.filter (fun x -> nan.Equals x |> not)
                let nRatios = rsOfProt |> Seq.length
                let nQuantifiedInCGE = cgeAbundances |> Seq.length
                let nQuantifiedInMCherry = mCherryAbundances |> Seq.length
                let meanR = 
                    if nRatios > 0 then
                        Seq.mean rsOfProt
                    elif nQuantifiedInCGE > 0 && nQuantifiedInMCherry = 0 then 
                        highestRatio 
                    elif nQuantifiedInMCherry > 0 && nQuantifiedInCGE = 0 then 
                        lowestRatio     
                    else nan           
                let stDevR = 
                    if nRatios > 1 then
                        Seq.stDev rsOfProt
                    else 
                        0.
                    
                let meanDiff = 
                    if nRatios > 0 then
                        Seq.mean dsOfProt
                    elif nQuantifiedInCGE > 0 && nQuantifiedInMCherry = 0 then 
                        highestRatio 
                    elif nQuantifiedInMCherry > 0 && nQuantifiedInCGE = 0 then 
                        lowestRatio     
                    else nan    
                let isCandidate = meanR > cutOff
                {|
                    Index           = k
                    // ProtName        = protName
                    IsCandidate     = isCandidate
                    MeanAbundance   = meanAbundance
                    MeanR           = meanR
                    StDevR          = stDevR
                    MeanD           = meanDiff
                    NRatios         = nRatios
                    NQuantifiedInCGE = nQuantifiedInCGE
                |}
            )
        |> Series.values
    stats
    |> Frame.ofRecords
    |> Frame.indexRows "Index"


let calculateStatsLog (f:Frame<string,string>) :Frame<string,string>=
    
    let r3s = 
        [calcRep f 1;calcRep f 2;calcRep f 3;calcRep f 4]
        |> Frame.mergeAll 

    let rCols = r3s.ColumnKeys |> Seq.filter (fun x -> x.Contains "r3s")
    let aCols = r3s.ColumnKeys |> Seq.filter (fun x -> x.Contains "abundance")
    let dCols = r3s.ColumnKeys |> Seq.filter (fun x -> x.Contains "diff_")
    let meanRNoSparse =  
        r3s
        |> Frame.dropSparseRows
        |> Frame.mapRows (fun k s -> 
                let rsOfProt = rCols |> Seq.map (fun ck -> s.GetAs<float>(ck)) |> Seq.map log10
                let meanR = Seq.mean rsOfProt
                meanR
            )
        |> Series.values
    
    let cutOff = 
        let borders =
            meanRNoSparse
            |> Array.ofSeq
            |> FSharp.Stats.Testing.Outliers.tukey 1.5
        borders.Upper

    let highestRatio :float = meanRNoSparse |> Seq.max
    let lowestRatio  :float = meanRNoSparse |> Seq.min
    let stats = 
        r3s
        |> Frame.mapRows (fun k s -> 
                let asOfProt = aCols |> Seq.map (fun ck -> s.GetAs<float>(ck,nan)) |> Seq.filter (fun x -> nan.Equals x |> not)
                let cgeAbundances = aCols |> Seq.filter (fun x -> x.Contains "cge") |> Seq.map (fun ck -> s.GetAs<float>(ck,nan)) |> Seq.filter (fun x -> nan.Equals x |> not)                
                let mCherryAbundances = aCols |> Seq.filter (fun x -> x.Contains "cge" |> not) |> Seq.map (fun ck -> s.GetAs<float>(ck,nan)) |> Seq.filter (fun x -> nan.Equals x |> not)                
                let meanAbundance = asOfProt |> Seq.map log10 |> Seq.mean
                let rsOfProt = rCols |> Seq.map (fun ck -> s.GetAs<float>(ck,nan)) |> Seq.filter (fun x -> nan.Equals x |> not) |> Seq.map log10
                let dsOfProt = dCols |> Seq.map (fun ck -> s.GetAs<float>(ck,nan)) |> Seq.filter (fun x -> nan.Equals x |> not) |> Seq.map log10
                let nRatios = rsOfProt |> Seq.length
                let nQuantifiedInCGE = cgeAbundances |> Seq.length
                let nQuantifiedInMCherry = mCherryAbundances |> Seq.length
                let meanR = 
                    if nRatios > 0 then
                        Seq.mean rsOfProt
                    elif nQuantifiedInCGE > 0 && nQuantifiedInMCherry = 0 then 
                        highestRatio 
                    elif nQuantifiedInMCherry > 0 && nQuantifiedInCGE = 0 then 
                        lowestRatio     
                    else nan           
                let stDevR = 
                    if nRatios > 1 then
                        Seq.stDev rsOfProt
                    else 
                        0.
                let meanDiff = 
                    if nRatios > 0 then
                        Seq.mean dsOfProt
                    elif nQuantifiedInCGE > 0 && nQuantifiedInMCherry = 0 then 
                        highestRatio 
                    elif nQuantifiedInMCherry > 0 && nQuantifiedInCGE = 0 then 
                        lowestRatio     
                    else nan   
                let isCandidate = meanR > cutOff
                {|
                    Index           = k
                    // ProtName        = protName
                    IsCandidate     = isCandidate
                    MeanAbundance   = meanAbundance
                    MeanR           = meanR
                    StDevR          = stDevR
                    MeanD           = meanDiff
                    NRatios         = nRatios
                    NQuantifiedInCGE = nQuantifiedInCGE
                |}
            )
        |> Series.values
    stats
    |> Frame.ofRecords
    |> Frame.indexRows "Index"

let plotRatiosVsMeanAbundance (f:Frame<string,string>) = 
    let dataAll = 
        f
        |> Frame.mapRows (fun k s -> 
            let a = s.GetAs<float>("MeanAbundance",nan)
            let r = s.GetAs<float>("MeanR",nan)
            let stDev = s.GetAs<float>("StDevR",nan)
            let l = ""//s.GetAs<string>("ProtName","")
            let c = s.GetAs<bool>("IsCandidate")
            {|
                    ProtName         = l
                    MeanAbundance    = a
                    MeanR            = r
                    StDevR           = stDev
                    IsCandidate      = c
            |}
            )
        |> Series.values
        |> Array.ofSeq
    
    let all = 
        let x = dataAll |> Seq.map (fun x -> x.MeanR) //|> Seq.map log10
        let y = dataAll |> Seq.map (fun x -> x.MeanAbundance) //|> Seq.map log10
        let label = dataAll |> Seq.map (fun x -> x.ProtName)
        let xError = dataAll |> Seq.map (fun x -> x.StDevR)
        // Chart.Point(x,y,"",Labels = label)
        Chart.Point(x,y)
        // |> Chart.withXErrorStyle (Array=xError,Symmetric=true)
        // |> Chart.withXErrorStyle(xError,Thickness = 0.5)
    let candidates =
        let dataFiltered = dataAll |> Array.filter (fun x -> x.IsCandidate)  
        let x = dataFiltered |> Seq.map (fun x -> x.MeanR) //|> Seq.map log10
        let y = dataFiltered |> Seq.map (fun x -> x.MeanAbundance) //|> Seq.map log10
        let label = dataFiltered |> Seq.map (fun x -> x.ProtName)
        let xError = dataFiltered |> Seq.map (fun x -> x.StDevR)
        Chart.Point(x,y)
        |> Chart.withXErrorStyle (Array=xError,Symmetric=true,Thickness=0.5,Width=3.)
        // |> Chart.withXErrorStyle(xError,Thickness = 0.5)    
    [
        all
        candidates
    ]
    |> Chart.combine
    |> Chart.withXAxisStyle "R3"
    |> Chart.withYAxisStyle "Mean intensity"


let plotHisto (f:Frame<string,string>) =
    let dataAll = 
        f
        |> Frame.mapRows (fun k s -> 
            let a = s.GetAs<float>("MeanAbundance",nan)
            let r = s.GetAs<float>("MeanR",nan)
            let stDev = s.GetAs<float>("StDevR",nan)
            let l = ""//s.GetAs<string>("ProtName","")
            let c = s.GetAs<bool>("IsCandidate")
            {|
                    ProtName         = l
                    MeanAbundance    = a
                    MeanR            = r
                    StDevR           = stDev
                    IsCandidate      = c
            |}
            )
        |> Series.values
        |> Array.ofSeq
    let histo =     
        let data = dataAll |> Seq.map (fun x -> x.MeanR) |> Seq.map log10  |> Seq.filter (fun x -> nan.Equals x |> not)|> Array.ofSeq
        let bw = FSharp.Stats.Distributions.Bandwidth.nrd0 data 
        let binned = FSharp.Stats.Distributions.Frequency.create  (bw) data |> Map.toSeq
        Chart.Column binned
        |> Chart.withXAxisStyle "R3"
    histo

let plotDiffVsMeanAbundance (f:Frame<string,string>) = 

    let dataAll = 
        f
        |> Frame.mapRows (fun k s -> 
            let a = s.GetAs<float>("MeanAbundance",nan)
            let r = s.GetAs<float>("MeanD",nan)
            let stDev = s.GetAs<float>("StDevR",nan)
            let l = ""//s.GetAs<string>("ProtName","")
            let c = s.GetAs<bool>("IsCandidate")
            {|
                    ProtName         = l
                    MeanAbundance    = a
                    MeanR            = r
                    StDevR           = stDev
                    IsCandidate      = c
            |}
            )
        |> Series.values
        |> Array.ofSeq
    
    let all = 
        let x = dataAll |> Seq.map (fun x -> x.MeanR) //|> Seq.map log10
        let y = dataAll |> Seq.map (fun x -> x.MeanAbundance) //|> Seq.map log10
        let label = dataAll |> Seq.map (fun x -> x.ProtName)
        let xError = dataAll |> Seq.map (fun x -> x.StDevR)
        // Chart.Point(x,y,"",Labels = label)
        Chart.Point(x,y)
        // |> Chart.withXErrorStyle (Array=xError,Symmetric=true)
        // |> Chart.withXErrorStyle(xError,Thickness = 0.5)
    let candidates =
        let dataFiltered = dataAll |> Array.filter (fun x -> x.IsCandidate)  
        let x = dataFiltered |> Seq.map (fun x -> x.MeanR) //|> Seq.map log10
        let y = dataFiltered |> Seq.map (fun x -> x.MeanAbundance) //|> Seq.map log10
        let label = dataFiltered |> Seq.map (fun x -> x.ProtName)
        let xError = dataFiltered |> Seq.map (fun x -> x.StDevR)
        Chart.Point(x,y)
        |> Chart.withXErrorStyle (Array=xError,Symmetric=true,Thickness=0.5,Width=3.)
        // |> Chart.withXErrorStyle(xError,Thickness = 0.5)    
    [
        all
        candidates
    ]
    |> Chart.combine
    |> Chart.withXAxisStyle "R3"
    |> Chart.withYAxisStyle "Mean intensity"

let res = calculateStats normalized

plotRatiosVsMeanAbundance res
|> Chart.show

plotHisto res
|> Chart.show

plotDiffVsMeanAbundance res
|> Chart.show

let resLog = calculateStatsLog normalized


plotRatiosVsMeanAbundance resLog
|> Chart.show

plotHisto resLog
|> Chart.show

plotDiffVsMeanAbundance resLog
|> Chart.show
///
let InPPerseus  = __SOURCE_DIRECTORY__ + @"\..\runs\CDJ_Perseus\MS_Perseus_CDJ5_final_onlyCDJ5.txt" 

///
let valuesPerseus :Frame<string,string> = 
    Frame.ReadCsv(path = InPPerseus, hasHeaders=true, separators="\t", schema="difference mittelwert = float")
    |> Frame.indexRows "T: Protein IDs" 
    // |> Frame.mapValues (fun x -> if x > 0. then x else nan)
    // |> Frame.sortRowsBy "Intensity" (fun k -> -k)
    // |> Frame.sliceCols ["difference mittelwert"]

valuesPerseus.Print()
valuesPerseus.ColumnKeys
|> Array.ofSeq

// let together = 
//     Frame.join JoinKind.Inner resLog valuesPerseus

// together.ColumnKeys
// |> Array.ofSeq
// together.RowCount

// together
// |> Frame.mapRows (fun k s ->s.GetAs<float>("MeanR"),s.GetAs<float>("difference mittelwert"))
// |> Series.values
// |> Seq.length

// together
// |> Frame.mapRows (fun k s -> s.GetAs<float>("MeanR"),s.GetAs<float>("difference mittelwert"))
// |> Series.values
// |> Chart.Point
// |> Chart.withXAxisStyle "meanR"
// |> Chart.withYAxisStyle "diff mittelwert"
// |> Chart.show


let together = 
    Frame.join JoinKind.Inner res valuesPerseus

together.ColumnKeys
|> Array.ofSeq
together.RowCount

together
|> Frame.mapRows (fun k s ->s.GetAs<float>("MeanD"),s.GetAs<float>("difference mittelwert"))
|> Series.values
|> Seq.length

together
|> Frame.mapRows (fun k s -> s.GetAs<float>("MeanD"),s.GetAs<float>("difference mittelwert"))
|> Series.values
|> Chart.Point
|> Chart.withXAxisStyle "meanR"
|> Chart.withYAxisStyle "diff mittelwert"
|> Chart.show
